@@ -6,10 +6,13 @@ from __future__ import annotations
 import logging
 import re
 import unicodedata
-from collections import OrderedDict
+from typing import Optional, TypedDict, Union
+from collections.abc import Sequence
 
-import dns
+import dns.exception
 import dns.resolver
+import dns.reversename
+from dns.nameserver import Nameserver
 import publicsuffixlist
 from expiringdict import ExpiringDict
 
@@ -44,6 +47,28 @@ MAILTO_REGEX = re.compile(MAILTO_REGEX_STRING, re.IGNORECASE)
 PSL = publicsuffixlist.PublicSuffixList()
 
 
+class NameserverResultOk(TypedDict):
+    hostnames: list[str]
+    warnings: list[str]
+
+
+class NameserverResultError(TypedDict):
+    hostnames: list[str]
+    error: str
+
+
+NameserverResult = Union[
+    NameserverResultOk,
+    NameserverResultError,
+]
+
+
+class MXHost(TypedDict):
+    hostname: str
+    preference: int
+    ip_addresses: list[str]
+
+
 class DNSException(Exception):
     """Raised when a general DNS error occurs"""
 
@@ -98,12 +123,12 @@ def query_dns(
     record_type: str,
     *,
     quoted_txt_segments: bool = False,
-    nameservers: list[str] = None,
-    resolver: dns.resolver.Resolver = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
     _attempt: int = 0,
-    cache: ExpiringDict = None,
+    cache: Optional[ExpiringDict] = None,
 ) -> list[str]:
     """
     Queries DNS
@@ -127,9 +152,9 @@ def query_dns(
     cache_key = f"{domain}_{record_type}_{quoted_txt_segments}"
     if cache is None:
         cache = DNS_CACHE
-    if type(cache) is ExpiringDict:
+    if isinstance(cache, ExpiringDict):
         records = cache.get(cache_key)
-        if records:
+        if isinstance(records, list):
             return records
     if not resolver:
         resolver = dns.resolver.Resolver()
@@ -212,8 +237,8 @@ def query_dns(
 def get_a_records(
     domain: str,
     *,
-    nameservers: list[str] = None,
-    resolver: dns.resolver.Resolver = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
 ) -> list[str]:
@@ -261,8 +286,8 @@ def get_a_records(
 def get_reverse_dns(
     ip_address: str,
     *,
-    nameservers: list[str] = None,
-    resolver: dns.resolver.Resolver = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
 ) -> list[str]:
@@ -306,9 +331,9 @@ def get_reverse_dns(
 def get_txt_records(
     domain: str,
     *,
-    nameservers: list[str] = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
     quoted_txt_segments: bool = False,
-    resolver: dns.resolver.Resolver = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
 ) -> list[str]:
@@ -354,11 +379,11 @@ def get_txt_records(
 def get_soa_record(
     domain: str,
     *,
-    nameservers: list[str] = None,
-    resolver: dns.resolver.Resolver = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
-) -> list[str]:
+) -> str:
     """
     Queries DNS for an SOA record
 
@@ -400,9 +425,9 @@ def get_soa_record(
 def get_nameservers(
     domain: str,
     *,
-    approved_nameservers: list[str] = None,
-    nameservers: list[str] = None,
-    resolver: dns.resolver.Resolver = None,
+    approved_nameservers: Optional[Sequence[str | Nameserver]] = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
 ) -> dict:
@@ -419,7 +444,7 @@ def get_nameservers(
         timeout_retries (int): The number of times to reattempt a query after a timeout
 
     Returns:
-        OrderedDict: A dictionary with the following keys:
+        dict: A dictionary with the following keys:
                      - ``hostnames`` - A list of nameserver hostnames
                      - ``warnings``  - A list of warnings
     """
@@ -444,28 +469,28 @@ def get_nameservers(
         raise DNSException(error)
 
     if approved_nameservers:
-        approved_nameservers = list(map(lambda h: h.lower(), approved_nameservers))
+        approved_nameservers = list(map(lambda h: str(h).lower(), approved_nameservers))
     for nameserver in ns_records:
         if approved_nameservers:
             approved = False
             for approved_nameserver in approved_nameservers:
-                if approved_nameserver in nameserver.lower():
+                if str(approved_nameserver).lower() in nameserver.lower():
                     approved = True
                     break
             if not approved:
                 warnings.append(f"Unapproved nameserver: {nameserver}")
 
-    return OrderedDict([("hostnames", ns_records), ("warnings", warnings)])
+    return {"hostnames": ns_records, "warnings": warnings}
 
 
 def get_mx_records(
     domain: str,
     *,
-    nameservers: list[str] = None,
-    resolver: dns.resolver.Resolver = None,
+    nameservers: Optional[Sequence[str | Nameserver]] = None,
+    resolver: Optional[dns.resolver.Resolver] = None,
     timeout: float = 2.0,
     timeout_retries: int = 2,
-) -> list[OrderedDict]:
+) -> list[MXHost]:
     """
     Queries DNS for a list of Mail Exchange hosts
 
@@ -478,7 +503,7 @@ def get_mx_records(
         timeout_retries (int): The number of times to reattempt a query after a timeout
 
     Returns:
-        list: A list of ``OrderedDicts``; each containing a ``preference``
+        list: A list of ``dicts``; each containing a ``preference``
                         integer and a ``hostname``
 
     Raises:
@@ -503,9 +528,7 @@ def get_mx_records(
             record = record.split(" ")
             preference = int(record[0])
             hostname = record[1].rstrip(".").strip().lower()
-            hosts.append(
-                OrderedDict([("preference", preference), ("hostname", hostname)])
-            )
+            hosts.append({"preference": preference, "hostname": hostname})
         hosts = sorted(hosts, key=lambda h: (h["preference"], h["hostname"]))
     except dns.resolver.NXDOMAIN:
         raise DNSExceptionNXDOMAIN("The domain does not exist.")
